/*
 * Proyecto:   Clase para uso de Excel a travs de OLE
 * Fichero:    TExcel.prg
 * Autor:      Bingen Ugaldebere  -  BiSoft
 * Fecha:      18/04/2018
 *
 * Basado en trabajos de varios usuarios de FiveWin
 *    Vctor Manuel Tomas Daz
 *    Daniel Andrade
 *    Rimantas Usevicius
 *    Carlos Sincur Romero
 *    Moiss Len
 *
 * More Info   https://msdn.microsoft.com/es-es/vba/vba-excel
 *
 *   xRange    utilizado en muchos mtodos puede ser un rango de una o varias celdas "A1"  o "B2:M3"
 *             o puede ser un array de posiciones numricas de 2 o 4 valores     {1,1}     o {2,2,3,13}   {fila, columna}
 */

#include "Xailer.ch"
#include "tExcel.ch"


CLASS TExcel

   Data oExcel
   Data oWorkBook
   Data oSheet
   Data oCell
   Data oWindow
   Data oShape
   Data oChart
   Data oClp

   Data cFileName          Init ""

   Data lInstalled         Init .F.
   Data cError             Init ""

   DATA cFont              Init "Segoe UI"
   DATA nFontSize          Init 10
   DATA lBold              Init .F.
   DATA lItalic            Init .F.
   DATA lUnderLine         Init .F.
   DATA nHAlign            Init 0
   DATA nVAlign            Init 1
   DATA lDefault           Init .F.

   DATA nFileFormat        Init xlOpenXMLWorkbook //.xlsx

   METHOD New()

   //Manejo de documentos Excel_________________________________________________
   METHOD NewDoc( cXLSName )
   METHOD Open( cFileName, cPassWord, lReadonly )
   METHOD IsReadOnly( lMsg )
   METHOD Protect( cPassword )               INLINE ::ProtectBook( cPassword ) , ::ProtectSheet( cPassword )
   METHOD UnProtect( cPassword )             INLINE ::UnProtectBook( cPassword ) , ::UnProtectSheet( cPassword )
   METHOD Save()
   METHOD SaveAs( cFileName , cPassWord )
   METHOD Close( lSave )                     INLINE ::oWorkBook:Close( lSave )
   METHOD End( lClose )

   //Manejo de libros dentro de Excel___________________________________________
   METHOD AddWorkBook()                                                          //Aade un libro de Excel nuevo
   METHOD DuplicateWorkBook()                INLINE ::oExcel:Sheets:Copy()       //Duplica un libro de Excel en otro nuevo
   METHOD CountBooks()                       INLINE ::oWorkBook:Count()          //Nos indica el nmero de libros abiertos
   METHOD ArrayOfBooks()                                                         //Devuelve un array con los nombre libros activos
   METHOD SeekBook( cBook )
   METHOD ProtectBook( cPassword )
   METHOD UnProtectBook( cPassword )         INLINE ::oWorkBook:UnProtect(  cPassword  )

   //Manejo de hojas dentro del libro actual____________________________________
   METHOD AddSheet( cSheetName, lLastSheet )
   METHOD DeleteSheet(cSheetName,lForce)
   METHOD MoveSheet( cSheetName, lAfter , nSheet)     //Pendiente
   METHOD SelectSheet(cSheetName)
   METHOD GetSheetName()                     INLINE ::oExcel:ActiveSheet:Name
   METHOD ReNameSheet(cSheetName,cNewName)   INLINE ::oExcel:Sheets(IF(cSheetName<>Nil,cSheetName,::oExcel:Sheets(::oExcel:ActiveSheet:Name):Name)):Name := cNewName
   METHOD CountSheets()                      INLINE ::oExcel:Sheets:Count()
   METHOD ArrayOfSheets()
   METHOD SeekSheet( cSheet )
   METHOD ProtectSheet( cPassword )
   METHOD UnProtectSheet( cPassword )        INLINE ::oSheet:UnProtect( cPassword  )

   //Visualizacin______________________________________________________________
   METHOD View( nView )                      //Muestra el libro de trabajo   nView = 1 Minimized 2 Normal 3 Maximized
   METHOD Hide()
   METHOD WebPagePreview()                   INLINE ::oWorkBook:WebPagePreview
   METHOD FreezePanes( lFreeze )             INLINE ::oWindow:FreezePanes := lFreeze

   //Previsualizacin de impresin e impresin__________________________________
   METHOD SetLandScape()                     INLINE ::oSheet:PageSetup:Orientation:=2
   METHOD SetPortrait()                      INLINE ::oSheet:PageSetup:Orientation:=1
   METHOD Headers( nOpc , cVal )
   METHOD Footers( nOpc , cVal )
   METHOD Margins( nOpc , nVal )
   METHOD SetPrintArea( xRange )
   METHOD lCenterH( lCenter)                 INLINE ::oSheet:PageSetup:CenterHorizontally:=lCenter
   METHOD lCenterV( lCenter)                 INLINE ::oSheet:PageSetup:CenterVertically:=lCenter
   METHOD Zoom( nZoom )                      INLINE ::oSheet:PageSetup:Zoom:=nZoom
   METHOD SetPage( nPage  )                  INLINE ::oSheet:PageSetup:PaperSize:=nPage
   METHOD TitleRows( xRange)
   METHOD Preview()
   METHOD Print()

   //Manejo de celdas dentro de la hoja actual del libro actual_________________
   METHOD SetPos( xRange )
   METHOD Goto( xRange )
   METHOD ColumnNumber( cColLetter )                        //Dndole una letra devuelve el nmero de columna  ColumnNumber( "AB" ) >> 28
   METHOD ColumnLetter( cColNumber )                        //Dndole un nmero devuelve el nmero de columna  ColumnLetter( 28 ) >> "AB"
   Method MakeRange( nRowIni, nColIni, nRowFin, nColFin )   //Mediante posiciones numricas devuelve un Rango de Excel
   Method aPosition( cCell )                 //             //Dndole una celda nos devuelve posiciones numricas
   METHOD GetCell()                          INLINE ::oCell := ::oExcel:ActiveCell
   METHOD Merge( xRange )                                   //Combinar celdas de un rango en una sola
   METHOD AutoFit( cColsRange )              INLINE ::oSheet:Columns( cColsRange ):AutoFit()   //Autoajusta las columnas indicadas  Autofit("A:H")
   METHOD ProtectCell( xRange )                            //Prepara la celda o rango de celdas para ser protegida/s si se protege la hoja de modificaciones
   METHOD UnProtectCell( xRange )                          //Quita la proteccin a la celda o rango de celdas

   //Manejo de filas y columnas dentro de la hoja actual del libro actual_______
   METHOD InsertRow( xRange )
   METHOD InsertColumn( xRange )

   Method RowHeight( xRange, nHeight )
   Method ColumnWidth( xRange, nWidth )

   METHOD HiddeRow( xRange )
   METHOD HiddeColumn( xRange )

   METHOD RowsCount()                        INLINE ::oSheet:UsedRange:Rows:Count()
   METHOD ColsCount()                        INLINE ::oSheet:UsedRange:Columns:Count()

   METHOD ColFormat( nCol , cFormat )        INLINE ::oSheet:Columns( nCol ):NumberFormat:=cFormat

   //Obtener datos de una hoja__________________________________________________
   METHOD Get( xRange )                                  //Devuelve un array con los valores del rango
   METHOD GetAll( lNoEmptyFiles, lTranspose  )            //Devuelve un array con el contenido de la pgina

   Method Transpose(aArray)                              //Transpone un array filas-columnas a columnas-filas

   //Poner datos en una hoja____________________________________________________
   METHOD Say( xRange , xValue, cFormat, nTextColor, nForeColor, nBorderStyle, nBorderColor, ;
               cFont, nFontSize, lBold, lItalic, lUnderline, nHAlign, nVAlign, nOrientation, lLocked, lShowZeros  )
   METHOD Say_Simple( xRange, xValue )
   METHOD CellFormat( xRange , cFormat, nTextColor, nForeColor, nBorderStyle, nBorderColor )
   METHOD CellFont( xRange , cFont, nFontSize, lBold, lItalic, lUnderline )
   METHOD CellFontBold( xRange, lBold )
   METHOD Borders( xRange , nBorderStyle, lTop, lLeft, lRight, lBottom, lInside )
   METHOD RangeColor( xRange , nTextColor, nForeColor, nBorderColor )

   METHOD AddComment( xCell , cText , nWidth , nHeight , lVisible )  //Aade un PostIt con un comentario visible o no

   METHOD AddPicture( xCell, cFile )                                 //Inserta una imagen en la celda indicada sin control del tamao
   METHOD AddPicture2( cFile, nTop , nLeft, nWidth , nHeight )       //Inserta una imagen en una posicin en pixel dndole un tamao en pixel

   METHOD Formula( xCell , cFormula )                                //Inserta una formula en la celda

   METHOD MultiLine( xRange )                                        //Convierte el texto de las celdas de un rango en multilinea

   METHOD Clear( xRange )                                            //Limpia el rango indicando
   METHOD Style( xRange, cStyle )                    Inline ::oSheet:Range( xRange ):Style := cStyle
   METHOD CellAligment( xRange, xHOrizontal, xVertical )


   //Seleccin de partes de la hoja_____________________________________________
   METHOD SelectAll()                        Inline   ::oSheet:UsedRange:Select()
   METHOD SelectRow(nLine)
   METHOD SelectColumn(xColumn)              //SelectColumn(3) == SelectColumn("C")
   METHOD UnSelect()

   //Cortar - Copiar y Pegar____________________________________________________
   METHOD Cut( xRange )
   METHOD Copy( xRange )
   METHOD CutAll()                           Inline   ::oSheet:UsedRange:Cut()
   METHOD CopyAll()                          Inline   ::oSheet:UsedRange:Copy()
   METHOD Paste()                            INLINE   ::oSheet:Paste()
   METHOD PasteSpecial( xRange, nPasteType )

   //Tratamiento de texto_______________________________________________________
   METHOD Find( cSearch, lMatch, lPart )
   METHOD FindNext( )
   METHOD Replace( cSearch, cReplace, lMatch, lPart, lAll, lFull, cFormat )
   METHOD CheckSpelling()

   //Graficacin
   METHOD Chart( xRange ,  nType , xTitleRange, xSeriesNameRange ,nDepth , nGapDepth )

   ERROR HANDLER ERROR()

ENDCLASS

//------------------------------------------------------------------------------

METHOD NEW()  CLASS TExcel

   TRY
      ::oExcel := GetActiveObject( "Excel.Application" )
   CATCH
      TRY
         ::oExcel    := CreateObject( "Excel.Application" )
         ::lInstalled := .T.
      CATCH
         ::lInstalled := .F.
         ::cError:="Excel no est instalado en este equipo"
         LogDebug( ::cError )
      END
   END

Return Self

METHOD ERROR() CLASS TExcel
   LOCAL cMsg, nParam
//   nParam := PCount()
   cMsg   := __GetMessage()
   MsgInfo( "La propiedad "+__GetMessage() +" no existe " , "Aviso al usuario")
RETURN Self

//Manejo de documentos Excel____________________________________________________

//Crea un nuevo documento Excel
METHOD NewDoc( cXLSName, cTemplate )  CLASS TExcel
   DEFAULT cXLSName  TO "LibroExcel1"
   DEFAULT cTemplate TO ""

   ::cFileName:=cXLSName

   ::AddWorkBook(,cTemplate)

   ProcessMessages(1000)

Return Nil

//Aade un libro de trabajo con una hoja a un documento Excel incializandolo
METHOD AddWorkBook( cXLSName, cTemplate )  CLASS TExcel
   DEFAULT cTemplate TO ""

   Try
      ::oWorkBook    := ::oExcel:Workbooks:Add(cTemplate)
      ::oSheet       := ::oWorkBook:WorkSheets( 1 )
      ::oWindow      := ::oExcel:ActiveWindow
      ::oShape       := ::oSheet:Shapes
      ::nFileFormat  := ::oWorkBook:FileFormat

      ::SetPos("A1")
      ::GetCell()

      ::oClp         := TClipboard():Create( )
   Catch
      ::lInstalled := .F.
      ::cError:= "ERROR! Excel worksheet not available. [" + Ole2TxtError()+ "]"
      LogDebug( ::cError )
      ::oExcel := NIL
      ::oForm:Close()
      RETURN .F.
   END
Return Nil


//Abre un documento Excel por su nombre
METHOD Open( cXLSName, cPassWord, lReadOnly ) CLASS TExcel
   Local lOk := .F.

   Default cXLSName  To ::cFileName
   DEFAULT cPassWord TO NIL
   Default lReadOnly To .F.

   If File( cXLSName )
      ::cFileName:=IF(At("\",cXLSName)>0,cXLSName,Application:cDirectory+cXLSName)

      Try
         ::oExcel:WorkBooks:Open( ::cFileName,3,lReadOnly,2,cPassWord )
      Catch
         ::oExcel:WorkBooks:Open( ::cFileName,3,lReadOnly,2 )
      End

      ::oWorkBook   := ::oExcel:ActiveWorkBook
      ::oSheet      := ::oWorkBook:WorkSheets( 1 )
      ::oWindow     := ::oExcel:ActiveWindow
      ::oShape      := ::oSheet:Shapes
      ::nFileFormat := ::oWorkBook:FileFormat

     ::SetPos("A1")
     ::GetCell()
      lOk := .T.
   else
      ::cError:= "No se ha encontrado el archivo Excel "+CRLF+CRLF+cXLSName
      MsgStop(::cError,"Error de apertura" )
      lOk := .F.
   endif

   ProcessMessages(1000)

Return lOk

//Graba un documento Excel
METHOD Save( ) CLASS TExcel

   If ::lOpen
      ::oWorkBook:Save( )
   Else
      ::SaveAs(  )
   Endif

   ProcessMessages(1000)

Return Nil

//Graba un documento Excel dndole un nombre y opcionalmente una contrasea
METHOD SaveAs(cFileName, cPassWord) CLASS TExcel
   DEFAULT cFileName To ::cFileName
   DEFAULT cPassWord TO NIL

   ::oWorkBook:SaveAs( cFileName , ::nFileFormat, cPassWord ,,,.T.,.T. )

   ProcessMessages(1000)

Return Nil

//Devuelve si el documento actual es de solo lectura
METHOD IsReadOnly(lMsg)  CLASS TExcel
Local lOk := .F.
   IF ::oBook:ReadOnly
      lOk := .T.
      IF lMsg
         MsgInfo(" El archivo " +::cFileName + " est abierto en otra sesin." )
      ENDIF
   ENDIF
RETURN lOk

METHOD ProtectBook( cPassword ) CLASS TExcel
   ::oWorkBook:Protect( cPassword , .T. , .T. )
RETURN( Nil )

METHOD ProtectSheet( cPassword ) CLASS TExcel
   ::oSheet:Protect( cPassword , .T. , .T. , .T. , .T. )
RETURN( Nil )


//Finaliza el documento y destruye el objeto Excel
METHOD End( lClose ) CLASS TExcel
  DEFAULT lClose  To .T.
  IF !lClose
    ::oExcel:WorkBooks:Close()
  ELSE
    ::oExcel:Quit()
  ENDIF

  ::oClp:End()

RETURN NIL


//Visualizacin_________________________________________________________________
METHOD View(nView)  CLASS TExcel
   Default nView To 2
   ::oExcel:WindowState := {xlMinimized,xlNormal,xlMaximized}[nView]  //1 Minimized 2 Normal 3 Maximized
   ::oExcel:Visible := .T.
   ::oExcel:Visible := .F.
   ::oExcel:Visible := .T.
Return Nil

METHOD Hide()  CLASS TExcel
   ::oExcel:Visible := .F.
Return Nil

//Previsualizacin de impresin e impresin_____________________________________

METHOD Headers( nOpc , cVal )  CLASS TExcel
DEFAULT cVal To 1  // Centrado
DEFAULT nOpc To 0
   DO CASE
      CASE nOpc == 1  // Center
           ::oSheet:PageSetup:CenterHeader:=cVal
      CASE nOpc == 2  // Left
           ::oSheet:PageSetup:LeftHeader:=cVal
      CASE nOpc == 3  // Right
           ::oSheet:PageSetup:RightHeader:=cVal
   ENDCASE
RETURN Nil

METHOD Footers( nOpc , cVal )  CLASS TExcel
DEFAULT cVal To 1  // Centrado
DEFAULT nOpc To 0
DO CASE
   CASE nOpc == 1  // Center
        ::oSheet:PageSetup:CenterFooter:=cVal
   CASE nOpc == 2  // Left
        ::oSheet:PageSetup:LeftFooter:=cVal
   CASE nOpc == 3  // Right
        ::oSheet:PageSetup:RightFooter:=cVal
ENDCASE
RETURN Nil

METHOD Margins( nOpc , nVal ) CLASS TExcel
DEFAULT nVal To 0
DEFAULT nOpc To 0
DO CASE
   CASE nOpc == 0  // Todos
        ::oSheet:PageSetup:RightMargin:=nVal
        ::oSheet:PageSetup:TopMargin:=nVal
        ::oSheet:PageSetup:LeftMargin:=nVal
        ::oSheet:PageSetup:BottomMargin:=nVal
        ::oSheet:PageSetup:FooterMargin:=nVal
        ::oSheet:PageSetup:HeaderMargin:=nVal
   CASE nOpc == 1  // Right
        ::oSheet:PageSetup:RightMargin:=nVal
   CASE nOpc == 2  // Top
        ::oSheet:PageSetup:TopMargin:=nVal
   CASE nOpc == 3  // Left
        ::oSheet:PageSetup:LeftMargin:=nVal
   CASE nOpc == 4  // Bottom
        ::oSheet:PageSetup:BottomMargin:=nVal
   CASE nOpc == 5  // Footer Margin
        ::oSheet:PageSetup:FooterMargin:=nVal
   CASE nOpc == 6  // Header Margin
        ::oSheet:PageSetup:HeaderMargin:=nVal
ENDCASE
RETURN Nil

METHOD SetPrintArea( xRange ) CLASS TExcel

   If Valtype(xRange)="A"
      xRange:=ASize(xRange,4)
      xRange:=::MakeRange( xRange[1], xRange[2], xRange[3], xRange[4] )
   Endif

   ::oSheet:PageSetup:PrintArea:=xRange

RETURN Nil

METHOD TitleRows( xRange) CLASS TExcel

   If Valtype(xRange)="A"
      xRange:=ASize(xRange,4)
      xRange:=::MakeRange( xRange[1], xRange[2], xRange[3], xRange[4] )
   Endif

   ::oSheet:PageSetup:PrintTitleRows:=xRange

Return Nil

METHOD Preview() CLASS TExcel

   ::oSheet:PrintPreview()
   If !::oExcel:Visible
      ::oExcel:Visible:=.T.
   Endif

Return nil

//Imprime el documento Excel
METHOD Print()   CLASS TExcel

  ::oSheet:PrintOut()

Return Nil


//Posicionamiento________________________________________________________________
METHOD SetPos( xRange ) CLASS TExcel

   If Valtype(xRange)="A"
      xRange:=ASize(xRange,4)
      xRange:=::MakeRange( xRange[1], xRange[2], xRange[3], xRange[4] )
   Endif

   Try      //Si no se puede seleccionar es por que esta en modo vista protegida
      ::oSheet:Range( xRange ):Select()
   End
   ::GetCell()
Return Nil

METHOD Goto( xRange ) CLASS TExcel

   If Valtype(xRange)="A"
      xRange:=ASize(xRange,4)
      xRange:=::MakeRange( xRange[1], xRange[2], xRange[3], xRange[4] )
   Endif

   ::SetPos( xRange )
Return nil

METHOD ColumnNumber( cColLetter ) CLASS TExcel
Return ::oSheet:Columns( cColLetter ):Column

METHOD ColumnLetter( cColNumber ) CLASS TExcel
Local cAbsPosition:=::oSheet:Cells( 1, cColNumber  ):Address
Return SubStr( cAbsPosition,2,RAt("$",cAbsPosition)-2)

METHOD Merge( xRange ) CLASS TExcel

   If xRange<>Nil
      If Valtype(xRange)="A"
         xRange:=ASize(xRange,4)
         xRange:=::MakeRange( xRange[1], xRange[2], xRange[3], xRange[4] )
      Endif

      ::oSheet:Range( xRange ):Merge()
   Endif

Return Nil

METHOD ProtectCell( xRange ) CLASS TExcel

   If xRange<>Nil
      If Valtype(xRange)="A"
         xRange:=ASize(xRange,4)
         xRange:=::MakeRange( xRange[1], xRange[2], xRange[3], xRange[4] )
      Endif

      ::oSheet:Range( xRange ):Locked := .T.
   Endif

Return Nil

METHOD UnProtectCell( xRange ) CLASS TExcel

   If xRange<>Nil
      If Valtype(xRange)="A"
         xRange:=ASize(xRange,4)
         xRange:=::MakeRange( xRange[1], xRange[2], xRange[3], xRange[4] )
      Endif

      ::oSheet:Range( xRange ):Locked := .F.
   Endif

Return Nil

//Manejo de filas y columnas dentro de la hoja actual del libro actual__________
METHOD InsertRow( xRange ) CLASS TExcel

   If xRange<>Nil
      If Valtype(xRange)="A"
         xRange:=ASize(xRange,4)
         xRange:=::MakeRange( xRange[1], xRange[2], xRange[3], xRange[4] )
      Endif

      ::SetPos( xRange )
   Endif

   ::GetCell():EntireRow:Insert()
Return Nil

METHOD InsertColumn( xRange ) CLASS TExcel

   If xRange<>Nil
      If Valtype(xRange)="A"
         xRange:=ASize(xRange,4)
         xRange:=::MakeRange( xRange[1], xRange[2], xRange[3], xRange[4] )
      Endif

      ::SetPos( xRange )
   Endif

   ::GetCell():EntireColumn:Insert()
Return Nil

Method RowHeight( xRange, nHeight ) CLASS TExcel

   If xRange<>Nil
      If Valtype(xRange)="A"
         xRange:=ASize(xRange,4)
         xRange:=::MakeRange( xRange[1], xRange[2], xRange[3], xRange[4] )
      Endif

      ::SetPos( xRange )
   Endif

   ::oSheet:Rows(1):RowHeight:=nHeight

Return Nil

Method ColumnWidth( xRange, nWidth ) CLASS TExcel

   If xRange<>Nil
      If Valtype(xRange)="A"
         xRange:=ASize(xRange,4)
         xRange:=::MakeRange( xRange[1], xRange[2], xRange[3], xRange[4] )
      Endif

      ::SetPos( xRange )
   Endif

   ::oSheet:Columns(Left(xRange,1)):ColumnWidth := nWidth

Return Nil

METHOD HiddeRow( xRange ) CLASS TExcel

   If xRange<>Nil
      If Valtype(xRange)="A"
         xRange:=ASize(xRange,4)
         xRange:=::MakeRange( xRange[1], xRange[2], xRange[3], xRange[4] )
      Endif

      ::SetPos( xRange )
   Endif

   ::GetCell():EntireRow():Hidden := .T.

Return Nil

METHOD HiddeColumn( xRange ) CLASS TExcel

   If xRange<>Nil
      If Valtype(xRange)="A"
         xRange:=ASize(xRange,4)
         xRange:=::MakeRange( xRange[1], xRange[2], xRange[3], xRange[4] )
      Endif

      ::SetPos( xRange )
   Endif

   ::GetCell():EntireColumn():Hidden := .T.

Return Nil


//Obtiene elementos de la hoja actual___________________________________________

//Devuelve el contenido de un rango
METHOD Get( xRange )  CLASS TExcel
   LOCAL xValues

   If Valtype(xRange)="A"
      xRange:=ASize(xRange,4)
      xRange:=::MakeRange( xRange[1], xRange[2], xRange[3], xRange[4] )
   Endif

   xValues := ::oSheet:Range( xRange ):Value

RETURN IF(ValType(xValues)="A",::Transpose(xValues),xValues)


//Devuelve el contenido de una pgina
METHOD GetAll( lNoEmptyLines, lTranspose )  CLASS TExcel
   Local xRange := ::MakeRange( 1, 1, ::RowsCount(),::ColsCount() )
   Local aDatos := ::oSheet:Range( xRange ):Value
   Local lVacio:=.F., nLine:=0, nCol:=0, aResult:={}

   Default lNoEmptyLines   to .F.
   Default lTranspose      to .F.

   If lNoEmptyLines
      For nLine:=1 to Len(aDatos)
         lVacio:=.F.
         For nCol := 1 To Len(aDatos[nLine])
            lVacio := aDatos[nLine, nCol]==Nil .Or. Len(AllTrim(ToString(aDatos[nLine, nCol])))=0
            If !lVacio
               AAdd(aResult,aDatos[nLine])
               Exit
            Endif
         Next
      Next
   Endif

RETURN IF(!lNoEmptyLines,IF(!lTranspose, aDatos, ::Transpose(aDatos) ),If(!lTranspose, aResult, ::Transpose(aResult) ) )


//Devuelve un array bidimensional transpuestas las filas y columnas
Method Transpose(aArray) CLASS TExcel
Local aTabla:={}, nFilas:=0,nColumnas:=0
Local aTrans:={}, nL:=0, nC:=0

   //Ver si es o no bidimensional
   IF ValType( aArray[ 1 ] ) = "A"     //Array de 2 dimensiones lo clono como est
      aTabla:=AClone(aArray)
   ELSE                          //Array de 1 dimensin no vale lo convierto en 2
      FOR nL := 1 TO Len( aArray )
         AAdd( aTabla, { aArray[ nL ] } )
      NEXT
   ENDIF

   nFilas:=Len(aTabla)
   nColumnas:=Len(aTabla[1])
   aTrans:=Array(nColumnas,nFilas)

   For nL:=1 to nFilas
      For nC:=1 to nColumnas
         aTrans[nC,nL]:=aTabla[nL,nC]
      Next
   Next

Return aTrans


//Say simple un rango := un valor
METHOD Say_Simple( xRange, xValue ) CLASS TExcel
   Default xRange       To "A1"
   ::oSheet:Range( xRange ):Value := xValue
RETURN NIL


//Say complejo con mltiples opciones
METHOD Say( xRange , xValue, cFormat, nTextColor, nForeColor, nBorderStyle, nBorderColor, cFont, nFontSize, lBold, lItalic, lUnderline, ;
            nHAlign, nValign, nOrientation, lLocked, lShowZeros  ) CLASS TExcel
Local xVal:="", aPosition:={0,0,0,0}, nRow:=1, nColumn:=1, cXlsData:=""

   Default xRange       To "A1"
   DEFAULT cFormat      To "0"               //
   Default nTextColor   To clBlack, nForeColor To clWhite, nBorderColor To clWhite
   DEFAULT nBorderStyle To 1
   DEFAULT cFont        To ::cFont
   DEFAULT nFontSize    To ::nFontSize
   DEFAULT lBold        To ::lBold
   DEFAULT lItalic      To ::lItalic
   DEFAULT lUnderline   To ::lUnderline
   DEFAULT nHAlign      To ::nHAlign        //0 xlHAlignGeneral 1 xlHAlignLeft 2 xlHAlignCenter 3 xlHAlignRight 4 xlHAlignJustify
   DEFAULT nValign      To ::nValign        //0 xlVAlignDistributed  1 xlVAlignTop 2 xlVAlignCenter  3 xlVAlignBottom  4 xlVAlignJustify
   DEFAULT nOrientation To 0
   DEFAULT lLocked      To .T.
   Default lShowZeros   To .F.


   If Valtype(xRange)="A"
      xRange:=ASize(xRange,4)
      xRange:=::MakeRange( xRange[1], xRange[2], xRange[3], xRange[4] )
   Endif

   ::CellFormat( xRange , cFormat, nTextColor, nForeColor, nBorderStyle, nBorderColor )
   ::CellFont( xRange , cFont, nFontSize, lBold, lItalic, lUnderline )

   Do case
      Case ValType(xValue)="C" .Or. ValType(xValue)="M"
         ::oSheet:Range( xRange ):Value               := xValue
      Case ValType(xValue)="N" .And. ( xValue<>0 .Or. (xValue=0 .And. lShowZeros) )
         ::oSheet:Range( xRange ):Value               := xValue
      Case ValType(xValue)="D"  .And. cFormat=="@"
         ::oSheet:Range( xRange ):Value               := StrTran(DToC(xValue),"-","/")
      Case ValType(xValue)="A"
         aPosition:=ASize(::aPosition(xRange),4)
         aPosition[3]:=aPosition[2]-1
         aPosition[4]:=aPosition[1]-1

         For nRow:=1 to Len(xValue)
            aPosition[4]++
            For nColumn:=1 To Len(xValue[1])
               If nRow=1
                  aPosition[3]++
               Endif
               Do case
               Case ValType(xValue[nRow,nColumn])="C"
                  cXlsData += xValue[nRow,nColumn]+Chr(9)
               Case ValType(xValue[nRow,nColumn])="N"
                  cXlsData += ForNum(xValue[nRow,nColumn],14,6)+Chr(9)
               OtherWise
                  cXlsData += ToString(xValue[nRow,nColumn])+Chr(9)
               End
            Next
            cXlsData := Left(cXlsData,Len(cXlsData)-1)+CRLF
         Next

         ::oClp:Empty()
         cXlsData := Left(cXlsData,Len(cXlsData)-1)
         ::SetPos( ::MakeRange( aPosition[1], aPosition[2], aPosition[4], aPosition[3] ) )
         ::oClp:SetText( cXlsData )
         ::Paste()
         ::oClp:Empty()
      OtherWise
         ::oSheet:Range( xRange ):Value               := xValue
   EndCase

   ::oSheet:Range( xRange ):HorizontalAlignment := {xlHAlignGeneral,xlHAlignLeft,xlHAlignCenter,xlHAlignRight,xlHAlignJustify}[nHAlign+1]
   ::oSheet:Range( xRange ):VerticalAlignment   := {xlVAlignDistributed,xlVAlignTop,xlVAlignCenter,xlVAlignBottom,xlVAlignJustify}[nValign+1]
   ::oSheet:Range( xRange ):Orientation         := nOrientation
   ::oSheet:Range( xRange ):Locked              := lLocked

Return Nil

METHOD CellFormat( xRange , cFormat, nTextColor, nForeColor, nBorderStyle, nBorderColor )  CLASS TExcel
   DEFAULT cFormat      To "0"
   Default nTextColor   To clBlack, nForeColor To clWhite, nBorderColor To clWhite
   DEFAULT nBorderStyle To 1

   If Valtype(xRange)="A"
      xRange:=ASize(xRange,4)
      xRange:=::MakeRange( xRange[1], xRange[2], xRange[3], xRange[4] )
   Endif

   ::oSheet:Range( xRange ):NumberFormat        := cFormat
   ::oSheet:Range( xRange ):Font:Color          := nTextColor
   ::oSheet:Range( xRange ):Interior:Color      := nForeColor
   ::oSheet:Range( xRange ):Borders():LineStyle := nBorderStyle
   ::oSheet:Range( xRange ):Borders:Color       := nBorderColor

Return Nil


METHOD CellFontBold( xRange, lBold ) CLASS TExcel

   DEFAULT lBold TO .F.

   If Valtype(xRange)="A"
      xRange:=ASize(xRange,4)
      xRange:=::MakeRange( xRange[1], xRange[2], xRange[3], xRange[4] )
   Endif

   ::oSheet:Range( xRange ):Font:Bold           := lBold

RETURN Nil

METHOD CellFont( xRange , cFont, nFontSize, lBold, lItalic, lUnderline )  CLASS TExcel
   DEFAULT cFont        To ::cFont
   DEFAULT nFontSize    To ::nFontSize
   DEFAULT lBold        To ::lBold
   DEFAULT lItalic      To ::lItalic
   DEFAULT lUnderline   To ::lUnderline

   If Valtype(xRange)="A"
      xRange:=ASize(xRange,4)
      xRange:=::MakeRange( xRange[1], xRange[2], xRange[3], xRange[4] )
   Endif

   ::oSheet:Range( xRange ):Font:Name           := cFont
   ::oSheet:Range( xRange ):Font:Size           := nFontSize
   ::oSheet:Range( xRange ):Font:Bold           := lBold
   ::oSheet:Range( xRange ):Font:Italic         := lItalic
   ::oSheet:Range( xRange ):Font:Underline      := lUnderline

Return Nil

METHOD Borders( xRange , nBorderStyle, nBorderColor, nBorderWeight, lTop, lLeft, lRight, lBottom, lInside )  CLASS TExcel

   Default nBorderStyle    to 1     //1 xlContinuous  2 xlDouble   3 xlDot   4 xlDash  5 xlDashDot  6 xlDashDotDot   7 xlLineStyleNone   8 xlSlantDashDot
   Default nBorderColor    To clBlack
   Default nBorderWeight   To 2     //1 xlThin  2 xlHairline 3 xlMedium  4 xlThick

   DEFAULT lTop         To .F., lLeft   To .F., lRight  To .F., lBottom To .F., lInside To .F.

   If Valtype(xRange)="A"
      xRange:=ASize(xRange,4)
      xRange:=::MakeRange( xRange[1], xRange[2], xRange[3], xRange[4] )
   Endif

   IF !lTop .AND. !lLeft .AND. !lRight .AND. !lBottom
      ::oSheet:Range( xRange ):Borders():LineStyle    := {xlContinuous, xlDouble, xlDot, xlDash, xlDashDot, xlDashDotDot, xlLineStyleNone, xlSlantDashDot}[nBorderStyle]
      ::oSheet:Range( xRange ):Borders:Color          := nBorderColor
      ::oSheet:Range( xRange ):Borders:Weight         := nBorderWeight
   ELSE
      IF lTop
         ::oSheet:Range( xRange ):Borders( xlEdgeTop ):LineStyle     := {xlContinuous, xlDouble, xlDot, xlDash, xlDashDot, xlDashDotDot, xlLineStyleNone, xlSlantDashDot}[nBorderStyle]
         ::oSheet:Range( xRange ):Borders( xlEdgeTop ):Color         := nBorderColor
         ::oSheet:Range( xRange ):Borders( xlEdgeTop ):Weight        := {xlThin, xlHairline, xlMedium, xlThick}[nBorderWeight]
      ENDIF
      IF lLeft
         ::oSheet:Range( xRange ):Borders( xlEdgeLeft ):LineStyle    := {xlContinuous, xlDouble, xlDot, xlDash, xlDashDot, xlDashDotDot, xlLineStyleNone, xlSlantDashDot}[nBorderStyle]
         ::oSheet:Range( xRange ):Borders( xlEdgeLeft ):Color        := nBorderColor
         ::oSheet:Range( xRange ):Borders( xlEdgeLeft ):Weight       := {xlThin, xlHairline, xlMedium, xlThick}[nBorderWeight]
      ENDIF
      IF lRight
         ::oSheet:Range( xRange ):Borders( xlEdgeRight ):LineStyle   := {xlContinuous, xlDouble, xlDot, xlDash, xlDashDot, xlDashDotDot, xlLineStyleNone, xlSlantDashDot}[nBorderStyle]
         ::oSheet:Range( xRange ):Borders( xlEdgeRight ):Color       := nBorderColor
         ::oSheet:Range( xRange ):Borders( xlEdgeRight ):Weight      := {xlThin, xlHairline, xlMedium, xlThick}[nBorderWeight]
      ENDIF
      IF lBottom
         ::oSheet:Range( xRange ):Borders( xlEdgeBottom ):LineStyle  := {xlContinuous, xlDouble, xlDot, xlDash, xlDashDot, xlDashDotDot, xlLineStyleNone, xlSlantDashDot}[nBorderStyle]
         ::oSheet:Range( xRange ):Borders( xlEdgeBottom ):Color      := nBorderColor
         ::oSheet:Range( xRange ):Borders( xlEdgeBottom ):Weight      := {xlThin, xlHairline, xlMedium, xlThick}[nBorderWeight]
      ENDIF
      IF lInside
         ::oSheet:Range( xRange ):Borders( xlInsideHorizontal ):LineStyle  := {xlContinuous, xlDouble, xlDot, xlDash, xlDashDot, xlDashDotDot, xlLineStyleNone, xlSlantDashDot}[nBorderStyle]
         ::oSheet:Range( xRange ):Borders( xlInsideHorizontal ):Color      := nBorderColor
         ::oSheet:Range( xRange ):Borders( xlInsideHorizontal ):Weight     := {xlThin, xlHairline, xlMedium, xlThick}[nBorderWeight]
         ::oSheet:Range( xRange ):Borders( xlInsideVertical ):LineStyle  := {xlContinuous, xlDouble, xlDot, xlDash, xlDashDot, xlDashDotDot, xlLineStyleNone, xlSlantDashDot}[nBorderStyle]
         ::oSheet:Range( xRange ):Borders( xlInsideVertical ):Color      := nBorderColor
         ::oSheet:Range( xRange ):Borders( xlInsideVertical ):Weight     := {xlThin, xlHairline, xlMedium, xlThick}[nBorderWeight]
      ENDIF
   ENDIF

RETURN Nil


METHOD RangeColor( xRange , nTextColor, nForeColor, nBorderColor )  CLASS TExcel

   Default nTextColor   To clBlack, nForeColor To clWhite, nBorderColor To clWhite

   If Valtype(xRange)="A"
      xRange:=ASize(xRange,4)
      xRange:=::MakeRange( xRange[1], xRange[2], xRange[3], xRange[4] )
   Endif

   ::oSheet:Range( xRange ):Font:Color          := nTextColor
   ::oSheet:Range( xRange ):Interior:Color      := nForeColor
   ::oSheet:Range( xRange ):Borders:Color       := nBorderColor
RETURN Nil


METHOD CellAligment( xRange, xHorizontal, xVertical )

   DEFAULT xHorizontal TO xlLeft
   DEFAULT xVertical   TO xlBottom

   If Valtype(xRange)="A"
      xRange:=ASize(xRange,4)
      xRange:=::MakeRange( xRange[1], xRange[2], xRange[3], xRange[4] )
   Endif

   ::oSheet:Range( xRange ):HorizontalAlignment  := xHorizontal
   ::oSheet:Range( xRange ):VerticalAlignment    := xVertical

RETURN NIL


METHOD AddComment( xCell , cText , nWidth , nHeight , lVisible )  CLASS TExcel

  DEFAULT cText      To ""
  DEFAULT nHeight    To 80, nWidth     To 80
  DEFAULT lVisible   To .F.

  IF !Empty( cText ) .And. xCell<>Nil
      If Valtype(xCell)="A"
         xCell:=ASize(xCell,4)
         xCell:=::MakeRange( xCell[1], xCell[2], xCell[3], xCell[4] )
      Endif

      With Object ::oSheet:Range( xCell ):AddComment(cText)
         :Shape:Height := nHeight
         :Shape:Width  := nWidth
         :Visible      := lVisible
      End
   Endif

RETURN Nil


METHOD Clear( xRange ) CLASS TExcel

   If Valtype(xRange)="A"
      xRange:=ASize(xRange,4)
      xRange:=::MakeRange( xRange[1], xRange[2], xRange[3], xRange[4] )
   Endif

  ::oExcel:Range( xRange ):Select()
  ::oExcel:Selection:ClearContents()

RETURN Nil


METHOD AddPicture( xCell, cFile ) CLASS TExcel

   If !Empty( cFile ) .And. File( cFile ) .And. xCell<>Nil
      If Valtype(xCell)="A"
         xCell:=ASize(xCell,4)
         xCell:=::MakeRange( xCell[1], xCell[2], xCell[3], xCell[4] )
      Endif

      ::SetPos( xCell )
      ::oSheet:Pictures:insert(cFile)
   Endif

RETURN Nil


METHOD AddPicture2( cFile, nTop , nLeft, nWidth , nHeight ) CLASS TExcel
  DEFAULT nTop    To 1, nLeft    to    1
  DEFAULT nWidth To 100 , nHeight To 100

  IF Empty( cFile ) .Or. !File( cFile )
     RETURN Nil
  ENDIF

  ::oShape:AddPicture( cFile , .T. , .T. , nLeft , nTop , nWidth , nHeight )

RETURN Nil


METHOD Formula( xCell, cFormula ) CLASS TExcel

   If xCell<>Nil
      If Valtype(xCell)="A"
         xCell:=ASize(xCell,4)
         xCell:=::MakeRange( xCell[1], xCell[2], xCell[3], xCell[4] )
      Endif

     ::oSheet:Range( xCell ):FormulaLocal:=IF(Left(AllTrim( cFormula ),1)<>"=","=","")+AllTrim( cFormula )
   Endif

RETURN Nil


METHOD MultiLine( xRange ) CLASS TExcel

   If xRange<>Nil
      If Valtype(xRange)="A"
         xRange:=ASize(xRange,4)
         xRange:=::MakeRange( xRange[1], xRange[2], xRange[3], xRange[4] )
      Endif

     ::oSheet:Range( xRange ):WrapText:=.T.
   Endif

Return Nil


//Tratamiento de texto_______________________________________________________
METHOD Find( cSearch, lMatch, lPart ) CLASS TExcel

  local oRange, lFound := .F.

  if cSearch == NIL
    RETURN lFound
  endif

  DEFAULT lMatch  To .F.  ,;
          lPart   To .F.

  ::GetCell():Activate()
  oRange := ::oSheet:Cells:Find( cSearch )

  if ValType( oRange ) == "O" .and. oRange[1] > 0
    oRange:Activate()

    ::GetCell()
    ::oFind   := oRange
    lFound    := .T.

    if (lMatch) .or. !(lPart)
      while !iif( lPart, cSearch $ ::Get( ::oCell:Row, ::oCell:Column ),  ;
                         cSearch == ::Get( ::oCell:Row, ::oCell:Column ))
        if !(::FindNext( oRange ))
          lFound  := .F.
          exit
        endif
      enddo
    endif
  endif

  RELEASE oRange

RETURN lFound


METHOD FindNext() CLASS TExcel

  local lFound := .F.
  local oRange, cRange, oCell

  if ValType( ::oFind ) == "O"
    oCell   := ::oCell
    cRange  := ::oExcel:ActiveCell:Address
    oRange  := ::oExcel:Cells:FindNext( ::oFind )

    if ValType( oRange ) == "O" .and. oRange[1] > 0
      oRange:Activate()
      ::GetCell()

      if ::oCell:Row == oCell:Row
        lFound  := ::oCell:Column > oCell:Column
      elseif ::oCell:Row > oCell:Row
        lFound  := .T.
      endif

      if lFound
        ::oFind := oRange
      else
        ::SetPos( cRange )
        ::oFind := NIL
      endif
    endif
  endif

RETURN lFound


METHOD Replace( cSearch, cReplace, lMatch, lPart, lAll, lFull, cFormat ) CLASS TExcel

  local lFound  := .F.

  DEFAULT lAll  To .F.
  DEFAULT lFull To .F.

  if cReplace != NIL
    while ::Find( cSearch, lMatch, lPart )
      lFound := .T.

      if cFormat != NIL
        ::CellFormat( ,,,, cFormat )
      endif

      if (lFull)
        ::Say(,, cReplace )
      else
        ::Say(,, StrTran(::Get(), cSearch, cReplace ) )
      endif

      if !(lAll)
        exit
      endif
    enddo
  endif

RETURN lFound


METHOD CheckSpelling() CLASS TExcel
   ::oSheet:CheckSpelling()
Return nil


//Seleccin de partes de la hoja________________________________________________
METHOD SelectRow( nRow ) CLASS TExcel
   ::oExcel:Sheets(::oExcel:Sheets(::oExcel:ActiveSheet:Name):Name):Rows(nRow):Select()
Return Nil

METHOD SelectColumn(nColumn) CLASS TExcel
   ::oExcel:Sheets(::oExcel:Sheets(::oExcel:ActiveSheet:Name):Name):Columns(nColumn):Select()
Return Nil

METHOD UnSelect() CLASS TExcel
   ::SetPos("A1")
Return Nil



//Cortar - Copiar y Pegar
METHOD Cut( xRange ) CLASS TExcel

   If xRange<>Nil
      If Valtype(xRange)="A"
         xRange:=ASize(xRange,4)
         xRange:=::MakeRange( xRange[1], xRange[2], xRange[3], xRange[4] )
      Endif

     ::oExcel:Range( xRange ):Select()
   Endif

  ::oExcel:Selection:Cut()

RETURN Self

METHOD Copy( xRange ) CLASS TExcel

   If xRange<>Nil
      If Valtype(xRange)="A"
         xRange:=ASize(xRange,4)
         xRange:=::MakeRange( xRange[1], xRange[2], xRange[3], xRange[4] )
      Endif

     ::oExcel:Range( xRange ):Select()
   Endif

  ::oExcel:Selection:Copy()

RETURN Nil

METHOD PasteSpecial( xRange, nPasteType ) CLASS TExcel

   Default nPasteType To 2 //1 xlPasteAll, 2 xlPasteFormulas, 3 xlPasteValues, 4 xlPasteFormats, 5 xlPasteComments, 6 xlPasteAllExceptBorders

   nPasteType:={xlPasteAll, xlPasteFormulas, xlPasteValues, xlPasteFormats, xlPasteComments, xlPasteAllExceptBorders}[nPasteType]

   If xRange<>Nil
      If Valtype(xRange)="A"
         xRange:=ASize(xRange,4)
         xRange:=::MakeRange( xRange[1], xRange[2], xRange[3], xRange[4] )
      Endif

     ::oExcel:Range( xRange ):Select( )
   Endif

   ::oExcel:Range( xRange ):PasteSpecial(nPasteType)

RETURN Nil



//Manejo de hojas dentro del libro actual_______________________________________

//Aadir hoja al libro actual
// cSheetName     [Opcional] Si se indica el nombre se le pone y si no ser asignado por Excel como HojaX
METHOD AddSheet( cSheetName, lLastSheet ) CLASS TExcel
Default lLastSheet To .T.

   ::oWorkBook:Sheets:Add()

   IF cSheetName<>Nil
      ::oExcel:Sheets(::oExcel:ActiveSheet:Name):Name := cSheetName
   Endif

   ::oSheet := ::oExcel:ActiveSheet

Return Nil

//Borrar una hoja del libro actual
// cSheetName     [Opcional] Si se indica el nombre se borra esa hoja y si no la hoja actual
// lForce         [Opcional] Si se indica .T.se borra la hoja sin preguntar
METHOD DeleteSheet(cSheetName,lForce) CLASS TExcel
Local lDisplayAlerts:=::oExcel:DisplayAlerts

   Default lForce To .F.

   IF lForce
      ::oExcel:DisplayAlerts:=.F.
   Endif

   ::oExcel:Sheets(IF(cSheetName<>Nil,cSheetName,::oExcel:Sheets(::oExcel:ActiveSheet:Name):Name)):Delete()

   ::oExcel:DisplayAlerts:=lDisplayAlerts

   ::oSheet := ::oExcel:ActiveSheet

Return Nil

//Mover una hoja del libro actual
// cSheetName     [Opcional] Si se indica el nombre se mueve esa hoja y si no la hoja actual
// lAfter         [Opcional] Si se indica .F. se mueve para la izquierda y si es .T. para la derecha

METHOD MoveSheet( cSheetName, lAfter , nSheet) CLASS TExcel

   Default lAfter To .T.
   Default nSheet To 1

  // cPos -> "After" | "Before"
*  METHOD MoveSheet(cSheet,cPos,nSheet)  INLINE ::oExcel:Sheets(cSheet):Move(cPos,nSheet)
*   oSheet := oExcel:Sheets(oSheet1)                              //move sheet position. This example will move
   ::oExcel:ActiveSheet:Move( "Hoja1" )
*::oExcel:Sheets( cSheetName ):Move( "Hoja1" )
   *MsgInfo(::oExcel:Sheets(IF(cSheetName<>Nil,cSheetName,::oExcel:Sheets(::oExcel:ActiveSheet:Name):Name)):Move(If(!lAfter,1,Nil),If(lAfter,1,Nil)))

   ::oSheet := ::oExcel:ActiveSheet

Return Nil

//Seleccionar  una hoja del libro actual
// cSheetName     Nombre de la hoja a Seleccionar [OBLIGATORIO]
METHOD SelectSheet(cSheetName) CLASS TExcel

   ::oExcel:Sheets(IF(cSheetName<>Nil,cSheetName,::oExcel:Sheets(::oExcel:ActiveSheet:Name):Name)):Select()

   ::oSheet := ::oExcel:ActiveSheet

Return Nil

//Devuelve un array de nombres de hojas del libro actual
METHOD ArrayOfSheets() CLASS TExcel
   LOCAL nItem:=1, aSheets := {}

   FOR nItem := 1 TO ::oExcel:Sheets:Count()
       Aadd( aSheets , ::oExcel:Sheets:Item( nItem ):Name )
   NEXT

RETURN aSheets

//Indica si existe la hoja dentro del libro actual
//Si existe devuelve su posicin y si no devuelve cero
METHOD SeekSheet( cSheetName ) CLASS TExcel
   Local aSheets  :=    ::ArrayOfSheets()
RETURN Ascan( aSheets , cSheetName )


Method aPosition( cCell )  CLASS TExcel
Local cChar:="", cRow:=""

   cCell:=Upper(cCell)

   For each cChar in cCell
      If Asc(cChar)>64 .And. Asc(cChar)<90
         cRow:=cRow+cChar
      Else
         Exit
      Endif
   Next

Return {::ColumnNumber( cRow ), Val(StrTran(cCell,cRow,""))}


Method MakeRange( nRowIni, nColIni, nRowFin, nColFin ) CLASS TExcel

  local cRange := ::ColumnLetter(nColIni) + AllTrim(Str(Int(nRowIni)))

  if nRowFin != NIL .and. nColFin != NIL
      cRange  += ":" + ::ColumnLetter(nColFin) + AllTrim(Str(Int(nRowFin)))
  endif

RETURN cRange


//Devuelve un array de libros en el Excel actual
METHOD ArrayOfBooks() CLASS TExcel
   LOCAL nItem:=1, aBooks := {}

   FOR nItem := 1 TO ::oWorkBook:Count()
       Aadd( aBooks , ::oWorkBook:Item( nItem ):Name )
   NEXT

RETURN aBooks

//Indica si existe el libro dentro del Excel actual
//Si existe devuelve su posicin y si no devuelve cero
METHOD SeekBook( cBookName ) CLASS TExcel
   Local aBooks  :=    ::ArrayOfBooks()
RETURN Ascan( aBooks , cBookName )



//Graficacin
METHOD Chart( xRange ,  nType , xTitleRange, xSeriesNameRange ,nDepth , nGapDepth ) CLASS TExcel
   LOCAL oSheet, nItem:=0
   DEFAULT xTitleRange  To "Grfica"
   Default xSeriesNameRange To {}
   DEFAULT nDepth       To 40       // Profundidad de la Grafica
   DEFAULT nGapDepth    To 20       // Separacion entre series
   Default nType        To xl3DColumn    // Barras 3D Consultar tipos en Excel.ch  Hay mas de 70

   If Valtype(xRange)="A"
      xRange:=ASize(xRange,4)
      xRange:=::MakeRange( xRange[1], xRange[2], xRange[3], xRange[4] )
   Endif

   If Valtype(xSeriesNameRange)="A"
      xSeriesNameRange:=ASize(xSeriesNameRange,4)
      xSeriesNameRange:=::MakeRange( xSeriesNameRange[1], xSeriesNameRange[2], xSeriesNameRange[3], xSeriesNameRange[4] )
   Endif

   Try      //Si no se puede seleccionar es por que esta en modo vista protegida
      ::oSheet:Range( xRange ):Select()
   End
   ::oExcel:Charts:Add()

   ::oChart                   := ::oExcel:ActiveChart
   ::oChart:ChartType         := nType

   ::oChart:ChartArea:Border:Weight      := xlThin // Negro
   ::oChart:ChartArea:Border:LineStyle   := xlContinuous
   ::oChart:PlotArea:Border:ColorIndex   := 1 // Negro
   ::oChart:PlotArea:Interior:ColorIndex := 2 // Blanco

   If Len(xSeriesNameRange)>0
      xSeriesNameRange:=::Get(xSeriesNameRange)
      ::oChart:HasLegend = .T.
      For nItem:=1 to Len(xSeriesNameRange)
         ::oChart:SeriesCollection(nItem):Name       := xSeriesNameRange[nItem,1]
      Next
   End

   ::oChart:HasTitle          := .T.
   ::oChart:ChartTitle:Text   := ToString(::Get(xTitleRange))
   ::oChart:DepthPercent      := nDepth
   ::oChart:GapDepth          := nGapDepth

RETURN Nil


//Ejemplos
/*
function Ejemplos_con_tExcel()

   With Object TExcel():New()
      :NewDoc()

      :Say( "A1", Date(),"@",clBlue,clYellow,,,"Tahoma",12,.t.,.t.,.t.,2,2,,1 )

      :Say( {1, 2}, Time(),"@",clGreen,clGray,,,"Tahoma",14,,,,2,2,,2 )

      :RowHeight( "A1", 120 )
      :ColumnWidth( "A1", 90 )
      :ColumnWidth( "B1", 100 )

      :Borders( "A5:H9",2,clBlue,1,.T.,.T.,.T.,.T.  )

      :AddComment( "C6", "Comentario de la celda C6" ,,,.T.)

      :AddPicture( "D1",Application:cDirectory+"Logotipo.Jpg" )

      :InsertRow( "A1" )
      :Say("B1","Ttulo centrado en 3 celdas combinadas",,clBlue,clLightGray,3,clBlue,"Tahoma",20,,,,2,2)
      :Merge("B1:D1")
      :RowHeight( "A1", 60 )

      :Say( "A12", 123 )
      :Say( "A13", 321 )

      :Formula( "A15","A12+A13")


      //Hoja nueva con datos y hoja para Grfica
      :AddSheet("Datos para la grfica")
      *:MoveSheet("Datos para la grfica",,1)
      :Say( "E1","Relacin de gastos e ingresos 2017",,clBlue,,,,"Tahoma",14,.T.,,,2,,,.T.)
      :Merge("E1:P1")
      :Say( "D3","Gastos")
      :Say( "D4","Ingresos")
      :Say( "B5", {{"Ene","Feb","Mar","Abr","May","Jun","Jul","Ago","Sep","Oct","Nov","Dic"}},)
      :Say( "C5", {{1,2,3,4,5,6,7,8,9,10,11,12},{11,12,13,14,15,16,17,18,19,20,21,22}},)

      :Chart("E3:P4",,"E1","D3:D4"  )

      //Visualizacin en maximizado
      :View(3)

      //Visualizacin en web
      *MsgInfo(:oExcel:ActiveWorkBook:Name)
*      :WebPagePreview()

*      Sleep(2000)
*      :Close(.f.)

      //Preview de impresin
      *:SetLandScape()
      *:Preview()

   End

Return Nil

*/

//ExcelCalendar(2020,"Horas del 2020")

Function ExcelCalendar( nYear, cTitle, aYearData, nType )
Local aMonthNames :={"Enero","Febrero","Marzo","Abril","Mayo","Junio","Julio","Agosto","Septiembre","Octubre","Noviembre","Diciembre"}
Local aDayNames   :={"Lunes","Martes","Mircoles","Jueves","Viernes","Sbado","Domingo",}
lOCAL nRow:=0, nCol:=0, nMonth:=0, nDay:= 0, dDay:=CToD("")
Local nPos:={{3,2},{3,17},{3,32},{3,47},{18,2},{18,17},{18,32},{18,47},{33,2},{33,17},{33,32},{33,47}}

   DEfault nYear     to Year(Date())
   DEfault cTitle    to "Ao "+ToString(nYear)
   Default aYearData To { {CToD("01-01-"+ToString(nYear)),"Ao nuevo"}, {CToD("06-01-"+ToString(nYear)),"Reyes"} }
   Default nType     To 1     //1 4x3 meses en la misma hoja, 2 3x4 en la misma hoja, 3 un mes por hoja


   With Object TExcel():New()
      :NewDoc(cTitle)
      :AddSheet( ToString(nYear),.T.)

      WaitOnMeter("Montando calendario Excel "+cTitle,,12)
      For nMonth:= 1 to 12
         AppData:oWaitOnMeter:nValue++
         nRow:=nPos[nMonth,1]; nCol:=nPos[nMonth,2]
         :Say({nRow,nCol},aMonthNames[nMonth],,clBlue,clLightGray,3,clBlue,"Tahoma",20,,,,2,2)
         :Merge({nRow,nCol,nRow,nCol+13})

         //Nombres de los das de la semana
         :Borders( {nRow++,nCol,nRow,nCol+13},2,clBlue,1,.T.,.T.,.T.,.T.  )
         For nDay:=1 to 7
            :Say({nRow,nCol+(nDay*2)-1},aDayNames[nDay],,clBlue,clLightGray,,,"Tahoma",12,,,,2,2)
         Next
         For nDay:=1 to 13 step 2
            :Merge({nRow,nCol+nDay-1,nRow,nCol+nDay})
         Next
         :Borders( {nRow++,nCol,nRow+11,nCol+13},2,clBlue,1,.T.,.T.,.T.,.T.  )

         //Pintar los das del mes
         dDay:=CToD("01-"+StrZero(nMonth,2)+"-"+ToString(nYear))
         Do while dDay<LastOfMonth(CToD("01-"+StrZero(nMonth,2)+"-"+ToString(nYear)))+1
            nDay:=IF(DoW(dDay)=1,7,DoW(dDay)-1)
            :Say({nRow,nCol+(nDay*2)-1},ToString(Day(dDay)),,IF(nDay=7,clRed,clBlack),,,,"Tahoma",12,,,,2,2)
            //Borde del da
            :Borders( {nRow,nCol,nRow+1,nCol+(nDay*2)-1},2,clBlue,1,.T.,.T.,.T.,.T.  )
            //Texto de los dias
            If AScan(aYearData,{|x| x[1]=dDay})>0
               LogDebug(AScan(aYearData,{|x| x[1]=dDay}))
               LogDebug(ToString(aYearData[AScan(aYearData,{|x| x[1]=dDay}),2]))
               :Say({nRow+1,nCol+(nDay*2)-2}, ToString(aYearData[AScan(aYearData,{|x| x[1]=dDay}),2]),,,,,,"Tahoma",12,,,,2,2)
            Endif
            :Merge({nRow+1,nCol+(nDay*2)-2,nRow+1,nCol+(nDay*2)-1})
            IF nDay=7
               nRow++
               nRow++
            Endif
            dDay++
         Enddo

      Next
      :Autofit("A:Z")
      WaitOff()
      :View(3) //Visualizacin en maximizado

   eND


Return .T.
